<?php

namespace InSegment\ApiCore\Api\V_2_0\MappingTypes;

use InSegment\ApiCore\Inspections\AutoCountInspection;
use InSegment\ApiCore\Models\TraverseState;
use InSegment\ApiCore\Models\SourceAttributeDownstream;
use InSegment\ApiCore\Facades\DTODefs;

/**
 * A template class for building custom mapping types
 */
abstract class MappingType
{
    /**
     * Source-Attribute-Downstream instance
     * 
     * @var \InSegment\ApiCore\Models\SourceAttributeDownstream
     */
    protected $nonRecursiveIterator;
    
    /**
     * An instance of version of class DTODefs
     * 
     * @var \InSegment\ApiCore\Api\V_2_0\DTOUtil
     */
    protected $dtoDefs;
    
    /**
     * Constructor
     * 
     * @param \InSegment\ApiCore\Models\SourceAttributeDownstream $nonRecursiveIterator
     */
    public function __construct(SourceAttributeDownstream $nonRecursiveIterator)
    {
        $this->dtoDefs = DTODefs::instance();
        $this->nonRecursiveIterator = $nonRecursiveIterator;
    }
    
    /**
     * Import
     * 
     * @param \InSegment\ApiCore\Models\TraverseState $state
     */
    public abstract function import(TraverseState $state);
    
    /**
     * Export
     * 
     * @param \InSegment\ApiCore\Models\TraverseState $state
     */
    public abstract function export(TraverseState $state);
    
    /**
     * Get information about external key name (and, optionally, value)
     * 
     * @param \InSegment\ApiCore\Api\V_1_0\DTOUtil $dtoDefs
     * @param array $mapping
     * @param array $data
     * @param array $param
     * @return array
     */
    public static function getExtKeysInfo($dtoDefs, $mapping, $data = null, array $param = []): array
    {
        $relatedClass = static::getRelatedClass($mapping);
        
        // for the 'find' type, prefer the specified key for search (maybe null) to default external key name
        $extKey = static::getKeyFromMapping($mapping) ?? $dtoDefs->getExternalKeyDefaultName($relatedClass);
        return $dtoDefs->getExtKeyInfo($relatedClass, $extKey, $data, $param);
    }
    
    /**
     * Get keys needed to be set for a successful search for a record by mapping
     * 
     * @param \InSegment\ApiCore\Api\V_2_0\DTOUtil $dtoDefs
     * @param mixed $mapping
     * @param mixed $baseRules
     * @param string $relatedClass
     * @param array $extKeysInfo
     * @return array
     */
    public static function getSearchAttributeKeys($dtoDefs, $mapping, $baseRules, $relatedClass, $extKeysInfo)
    {
        if (isset($mapping['uniqueSupply'])) {
            $convertedRules = $dtoDefs->rulesByConversion($mapping['uniqueSupply'], $baseRules);

            if (!is_array($convertedRules)) {
                $suppliesForUniques = null;
            } else {
                $suppliesForUniques = [];
                $conversionCount = count($convertedRules);
                foreach ($convertedRules as $dotFormat => $setMapping) {
                    if (is_null($setMapping)) {
                        continue;
                    }
                    
                    if ($dotFormat === '-' && $conversionCount === 1) {
                        $suppliesForUniques = null;
                        break;
                    }
                    
                    array_set($suppliesForUniques, $dotFormat, null);
                }
            }
        } else {
            $suppliesForUniques = [];
        }
        
        return array_keys($dtoDefs->uniquesToAttributes($suppliesForUniques, $relatedClass, $extKeysInfo));
    }
    
    /**
     * Should return an array of strings representing types of mappings
     * which are to be handled by this class
     * 
     * @return array
     */
    public abstract static function coversTypes(): array;
    
    /**
     * Enumerate relation type, return string with full new address if enumeration is to recurse
     * 
     * @param array $output
     * @param string $address
     * @param string $class
     * @param string $dotFormat
     * @param mixed $mapping
     * @return bool
     */
    public static function enumerateRelationType(&$output, $address, $class, $dotFormat, &$mapping): bool
    {
        return static::importsRelation($mapping);
    }
    
    /**
     * Get relation name from mapping
     * 
     * @param string $class
     * @param array $mapping
     * @return null
     */
    public static function getRelationName($class, $mapping)
    {
        $relationName = array_get($mapping, 'relation');
        return $relationName && method_exists(new $class, $relationName) ? $relationName : null;
    }
    
    /**
     * Get related class for relation of mapping type
     * 
     * @param array $mapping
     * @return null
     */
    public static function getRelatedClass($mapping)
    {
        return null;
    }
    
    /**
     * Get class name of AutoCountInspection
     * 
     * @return string
     */
    public static function getAutoCountInspectionClass(): string
    {
        return AutoCountInspection::class;
    }
    
    /**
     * Whether this MappingType relation should look for an data to process its object
     * 
     * @param array $mapping
     * @return bool
     */
    public static function importsRelation($mapping): bool
    {
        return false;
    }
    
    /**
     * Check whether the mapping type is auto-countable (one-to-one with parent)
     * 
     * @param array $mapping
     * @return bool
     */
    public static function isAutoCount($mapping): bool
    {
        return false;
    }
    
    /**
     * Get key from mapping
     * 
     * @param array $mapping
     * @return array|string|null
     */
    protected static function getKeyFromMapping($mapping)
    {
        return null;
    }
    
    /**
     * Check whether records from the table of the related model could be deleted
     * 
     * @param array $mapping
     * @return bool
     */
    public static function isDeletionEnabled($mapping): bool
    {
        return false;
    }
    
}
